body {
  margin: 0;
  padding: 8px;
  font-family: Arial, "Helvetica Neue", Helvetica, sans-serif;
}
h1 {
  margin: 4px 0 4px 0;
}
img {
  border: 0;
  padding: 0 2px;
  vertical-align: text-bottom;
}
table.dir-listing {
  border: 0;
  border-collapse: collapse;
}
table.dir-listing tbody tr:hover
{
  background-color: #eee;
}
table.dir-listing tbody tr td a
{
  color: #000;
  text-decoration: none;
}
table.dir-listing tbody tr:hover td a
{
  text-decoration: underline;
}
table.dir-listing tr.directory td a
{
  font-weight: bold;
}
table.dir-listing th,
table.dir-listing td
{
  padding: 2px 20px 2px 2px;
  /* text-align: left; */
  font-family: monospace;
  vertical-align: bottom;
  /* white-space: pre; */
}
table.dir-listing th:nth-child(3),
table.dir-listing td:nth-child(3)
{
  text-align: right;
}
a.symlink {
  font-style: italic;
}
p.auth-user {
  font-size: smaller;
  text-align: right;
}
p.trailer {
  font-size: smaller;
}
